/*******************************************************************************
 * Copyright 2013 butor.com
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.butor.ldap;
import java.io.StringWriter;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

import javax.naming.AuthenticationException;
import javax.naming.Context;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.PartialResultException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;

import org.butor.utils.ApplicationException;
import org.butor.utils.CommonMessageID;

import com.google.common.base.Strings;

public class DefaultLdapUserModel implements LdapUserModel {
	private String ldapUrl = null;
	private String domain = null;
	private String searchBase = null;

	private String adminUsername = null;
	private String adminPwd = null;
	private DirContext ctx = null;

	public DefaultLdapUserModel(String ldapUrl, String domain) {
		super();
		this.ldapUrl = ldapUrl;
		this.domain = domain;
	}

	private DirContext getContext(String principal, String pwd) throws NamingException {
		// Set up the environment for creating the initial context
		Hashtable<String, String> env = new Hashtable<String, String>();
		env.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");

		env.put(Context.PROVIDER_URL, ldapUrl);
		if (ldapUrl.toLowerCase().startsWith("ldaps")) {
			env.put(Context.SECURITY_PROTOCOL, "ssl");
			env.put("java.naming.ldap.factory.socket", NoCheckSocketFactory.class.getName());
		}
		env.put(Context.SECURITY_AUTHENTICATION, "simple");
		env.put(Context.SECURITY_PRINCIPAL, principal);
		env.put(Context.SECURITY_CREDENTIALS, pwd);

		// Create context
		return new InitialDirContext(env);
	}

	@Override
	public void close() {
		if (ctx != null)
			try {
				ctx.close();
				ctx = null;
			} catch (NamingException e) {
				// Never mind this.
			}
	}
	@Override
	public LdapUser auth(String username, String pwd) {
		try {
			String id = username;
			String email = username;
			int pos = id.indexOf("@");
			if (pos == -1)
				email += "@" +domain;
			else
				id = id.substring(0,pos);

			ctx = getContext(email, pwd);
			if (ctx == null)
				return null;

			// http://docs.oracle.com/cd/E26217_01/E35769/html/ldap-filters-attrs-users.html#ldap-filters-attributes-users-default
			String searchFilter = "(&(objectClass=user)(sAMAccountName=" + id + "))";
			StringWriter swDomain = new StringWriter();
			String[] domainSplit = domain.split("\\.");
			for (int i = 0; i < domainSplit.length; i++) {
				swDomain.append(String.format("DC=%s", domainSplit[i]));
				if (i + 1 < domainSplit.length)
					swDomain.append(",");
			}
			SearchControls searchControls = new SearchControls();
			searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
			NamingEnumeration<SearchResult> results = ctx.search(swDomain.toString(), searchFilter, searchControls);

			LdapUser user = null;
			while (results.hasMoreElements()) {
				SearchResult sr = (SearchResult) results.next();
				Attributes attrs = sr.getAttributes();
				user = retrieveUser(attrs, true);
				break;
			}
			return user;
		} catch (AuthenticationException e) {
			// Never mind this.
		} catch (NamingException e) {
			throw new RuntimeException(e);
		} finally {
		}
		return null;
	}

	private LdapUser retrieveUser(Attributes attrs, boolean noNulls) throws NamingException {
		/*debug
		NamingEnumeration<? extends Attribute> attributes = attrs.getAll();
		while (attributes.hasMoreElements()) {
			Attribute attr = attributes.next();
			System.out.println(attr.getID() + ": " + attr.get());
		}
		*/
		Object sn = attrs.get("sn");
		Object givenName = attrs.get("givenName");
		Object mail = attrs.get("mail");
		Object displayName = attrs.get("displayName");
		Object sAMAccountName = attrs.get("sAMAccountName");

		if (noNulls && (sn == null || givenName == null ||
				displayName == null)) {
			return null;
		}
		LdapUser user = new LdapUser();
		user.setUsername(sAMAccountName == null ? null :  (String)((Attribute)sAMAccountName).get());
		user.setLastName(sn == null ? null :  (String)((Attribute)sn).get());
		user.setFirstName(givenName == null ? null : (String)((Attribute)givenName).get());
		user.setEmail(mail == null ? null : (String)((Attribute)mail).get());
		user.setDisplayName(displayName == null ? null : (String)((Attribute)displayName).get());
		return user;
	}
	@Override
	public List<LdapUser> search(String name) {
		return search(name,0);
	}
	public List<LdapUser> search(String name, int retriesSoFar) {
		if (Strings.isNullOrEmpty(this.searchBase)) {
			ApplicationException.exception(CommonMessageID.MISSING_CONFIG.getMessage("searchBase"));
		}
		if (Strings.isNullOrEmpty(this.adminUsername)) {
			ApplicationException.exception(CommonMessageID.MISSING_CONFIG.getMessage("adminUsername"));
		}
		if (Strings.isNullOrEmpty(this.adminPwd)) {
			ApplicationException.exception(CommonMessageID.MISSING_CONFIG.getMessage("adminPwd"));
		}

		NamingEnumeration<?> results = null;
		try {
			if (ctx == null) {
				ctx = getContext(adminUsername, adminPwd);
			}
			if (ctx == null) {
				return null;
			}

			LinkedList<LdapUser> list = new LinkedList<LdapUser>();

			SearchControls controls = new SearchControls();
			controls.setSearchScope(SearchControls.SUBTREE_SCOPE);
			controls.setReturningAttributes(new String[]{
					"sAMAccountName", "sn", "givenName",
					"mail", "displayName"});
			String filter = (Strings.isNullOrEmpty(name) ? "(objectclass=user)" : 
						String.format("(&(objectClass=user)(|(userPrincipalName=*%s*)(mail=*%s*)))", name, name));
			results = ctx.search(searchBase, filter, controls);

			while (true) {
				try {
					if (results.hasMore()) {
						SearchResult sr = (SearchResult) results.next();
						LdapUser user = retrieveUser(sr.getAttributes(), true);
						if (user != null) {
							list.add(user);
						}
					} else {
						break;
					}
				} catch (PartialResultException e) {
					//OK no more result
					break;
				}
			}
			return list;

		} catch (NameNotFoundException e) {
			// The base context was not found.
			// Just clean up and exit.
			e.printStackTrace();

		} catch (NamingException e) {
			if (ctx != null && retriesSoFar < 2) {
				close();
				return search(name, retriesSoFar+1);
			}
			throw new RuntimeException(e);
		} finally {
			if (results != null) {
				try {
					results.close();
				} catch (NamingException e) {
					// Never mind this.
				}
			}
		}
		return null;
	}

	public void setSearchBase(String searchBase) {
		this.searchBase = searchBase;
	}

	public void setAdminUsername(String adminUsername) {
		this.adminUsername = adminUsername;
	}

	public void setAdminPwd(String adminPwd) {
		this.adminPwd = adminPwd;
	}

}
