/*******************************************************************************
 * Copyright 2013 butor.com
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.butor.mail;

import java.util.Collections;
import java.util.List;
import java.util.Properties;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import javax.mail.Address;
import javax.mail.Message;
import javax.mail.Message.RecipientType;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import org.butor.utils.StringUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author sawanai
 */
public final class Mailer implements IMailer {
	private Logger _logger = LoggerFactory.getLogger(this.getClass());
	private Queue<Message> _queue = new ConcurrentLinkedQueue<Message>();
	private Session session = null;
	private String smtpHost = null;
	private String smtpPort = "0";
	private InternetAddress _fromRecipient = null;
	private ScheduledExecutorService _executor = null;
	
	public Mailer(ScheduledExecutorService executor_) throws IllegalArgumentException, AddressException {
		_executor = executor_;

		_executor.scheduleWithFixedDelay(getMonitor(),
				5000, 5000, TimeUnit.MILLISECONDS);
	}
	protected Runnable getMonitor() {
		return new Runnable() {
			@Override
			public void run() {
				try {
					while (true) {
						// will stop when the queue is empty.
						Message message = getMessage();
						if (message == null)
							break;

						String adrs = "";
						for (Address adr : message.getRecipients(RecipientType.TO)) {
							adrs += adr.toString() +";";
						}
						_logger.info("sending mail to {}, subject={}", adrs, message.getSubject());
						// Send our mail message
						Transport.send(message);
					}
				} catch (Exception e) {
					_logger.error("Failed", e);
				}
			}
		};
	}
	@Override
	public void sendMail(String to, String subject, String content, String fromRecipient) {
		sendMail(Collections.singletonList(to), subject, content, fromRecipient);
	}
	
	@Override
	public void sendMail(List<String> to, String subject, String content, String fromRecipient) {
		try {
			if (session == null) {
				//TODO factory
				Properties props = new Properties();
				props.setProperty("mail.smtp.host", smtpHost);
				props.setProperty("mail.smtp.port", smtpPort);

				/*TODO
				if (session != null)
					session.close();
		*/
				session = Session.getDefaultInstance(props);
			}
			if (StringUtil.isEmpty(smtpHost)) {
				throw new IllegalArgumentException("smtp host not set!");
			}
			if (StringUtil.isEmpty(smtpPort)) {
				throw new IllegalArgumentException("smtp port not set!");
			}
			if (StringUtil.isEmpty(fromRecipient) && _fromRecipient == null) {
				throw new IllegalArgumentException("Missing from recipient arg!");
			}

			InternetAddress ia = _fromRecipient;
			if (fromRecipient != null)
				ia = new InternetAddress(fromRecipient);

			// Prepare our mail message
			Message message = new MimeMessage(session);
			message.setFrom(ia);

			InternetAddress dests[] = new InternetAddress[to.size()];
			int i = 0;
			for (String tt : to) {
				dests[i++] = new InternetAddress(tt);
			}

			message.setRecipients(Message.RecipientType.TO, dests);
			message.setSubject(subject);
			message.setContent(content, "text/plain; charset=UTF-8");

			_logger.info("queued mail to {}, subject={}", to.toString(), subject);
			_queue.add(message);
		} catch (Exception e) {
			_logger.error("Failed", e);
		}
	}

	@Override
	public Message getMessage() {
		return _queue.poll();
	}
	public String getSmtpHost() {
		return smtpHost;
	}
	public String getSmtpPort() {
		return smtpPort;
	}
	public String getFromRecipient() {
		if (_fromRecipient != null)
			return _fromRecipient.getAddress();
		return null;
	}
	public void setFromRecipient(String fromRecipient_) throws AddressException {
		_fromRecipient = new InternetAddress(fromRecipient_);
	}
	public void setSmtpHost(String smtpHost) {
		this.smtpHost = smtpHost;
	}
	public void setSmtpPort(String smtpPort) {
		this.smtpPort = smtpPort;
	}
}
